use std::process::{Command};
use std::io::{stdin,stdout,Write};
use colored::Colorize;
use regex::Regex;
use url::{Url, ParseError};

fn clear_screen() {
    println!("clearing");
    Command::new("clear")
        .status()
        .expect("Failed to clear screen");
}

fn parse_markdown(page_content: String) -> (String, Vec<String>) {
	let mut parsed_page_content: String = "".to_string();
	let mut hyperlink_number_counter: u64 = 0;
	let mut links: Vec<String> = Vec::new();
	let (screen_width, _screen_height) = termion::terminal_size().unwrap(); // So the horizontal line (<hr/>) spans the whole console
	
    for line in page_content.lines() {
        let mut parsed_line: String = line.to_string();
        // Bold
        let bold_regex = Regex::new(r"((\*\*)|(__))(.*?)((\*\*)|(__))").unwrap();
        parsed_line = bold_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            caps[4].bold().to_string()
        }).to_string();

        // Strikethrough
        let strikethrough_regex = Regex::new(r"~~(.*?)~~").unwrap();
        parsed_line = strikethrough_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            caps[1].strikethrough().to_string()
        }).to_string();

        // Horizontal lines
        let hr_regex = Regex::new(r"^(\*\*\*)|(---)|(___)$").unwrap();
        parsed_line = hr_regex.replace_all(&parsed_line, |_caps: &regex::Captures| {
            let mut result: String = "\n".to_string();
            for _x in 0..screen_width/2 {
                result += "- ";
            }
            result += "\n";
            result
        }).to_string();

        // html br tag support
        let br_regex = Regex::new(r"(.*?)<br/>(.*?)").unwrap();
        parsed_line = br_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            format!("{}{}{}", &caps[1], "\n", &caps[2])
        }).to_string();

        // Italics
        let italic_regex = Regex::new(r"\*(.*?)\*").unwrap();
        parsed_line = italic_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            caps[1].italic().to_string()
        }).to_string();
        let italic_regex = Regex::new(r"_(.*?)_").unwrap();
        parsed_line = italic_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            caps[1].italic().to_string()
        }).to_string();

        // Block quotes
        let block_quotes_regex = Regex::new(r"^>(.*)").unwrap();
        parsed_line = block_quotes_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            format!("   | {}", &caps[1])
        }).to_string();

        // Ordered list
        let ordered_list_regex = Regex::new(r"^([ \t]+|^)([0-9]+)\. (.*)").unwrap();
        parsed_line = ordered_list_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            format!("{} {}. {}", &caps[1], &caps[2], &caps[3])
        }).to_string();

        // Unordered list ([    ]+|^)- (.*)
        let unordered_list_regex = Regex::new(r"^([ \t]+|^)(-|\+|\*).(.*)").unwrap();
        parsed_line = unordered_list_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            format!("{} • {}", &caps[1], &caps[3])
        }).to_string();

        // Inline code
        let inline_code_regex = Regex::new(r"`([^`]+?)`").unwrap();
        parsed_line = inline_code_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            format!("{}", &caps[1].magenta())
        }).to_string();

        // HyperLink
        let hyperlink_regex = Regex::new(r"(.*?)\[(.*?)\]\((.*?)\)").unwrap();
        parsed_line = hyperlink_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            // Check if the character before the link is not '!'
            if !caps[1].ends_with('!') {  // caps[1] is everything before the link
                let result = format!("{}{}[{}]", &caps[1], &caps[2].blue().underline(), hyperlink_number_counter);
                let url = caps[3].to_string();
                links.push(url);
                hyperlink_number_counter += 1;
                result
            } else {
                // If it's an image (starts with !), return the link as is
                let url = caps[3].to_string();
                links.push(url);
                hyperlink_number_counter += 1;
                format!("({})[{}]", &caps[2].green(), hyperlink_number_counter)
            }
        }).to_string();
        
        let quick_hyperlink_regex = Regex::new(r"<(.*:\/\/.*)>").unwrap();
        parsed_line = quick_hyperlink_regex.replace_all(&parsed_line, |caps: &regex::Captures| {
            hyperlink_number_counter += 1;
            let url = caps[1].to_string();
            links.push(url);
            format!("{}[{}]", &caps[1].blue().underline(), hyperlink_number_counter)
        }).to_string();


        parsed_page_content+=&(parsed_line + "\n");
    }

    // multiline code
    let multiline_code_regex = Regex::new(r"(?ms)```((.*?\n)+?)```").unwrap();
    parsed_page_content = multiline_code_regex.replace_all(&parsed_page_content, |caps: &regex::Captures| {
        // Capture the code inside the %% blocks
        let code_block = &caps[1];
        
        // Add a tab to each line in the block
        let indented_code = code_block
            .lines()
            .map(|line| format!("\t{}", line)) // Insert tab at the start of each line
            .collect::<Vec<String>>()
            .join("\n");

        // Return the formatted block with magenta color
        format!("{}", indented_code.magenta())
    }).to_string();

    return (parsed_page_content, links);
}

fn fetch_page(url: &Url) -> String {
    let full_url_formatted = format!("{}", url);

    // Call curl using Com, mand
    let output = Command::new("curl")
        .arg(full_url_formatted)
        .output()
        .expect("Failed to execute curl command");

    // Check if the command was successful
    if output.status.success() {
        let page: String = String::from_utf8_lossy(&output.stdout).to_string();
        return page
    } else {
        eprintln!("Error:\n{}", String::from_utf8_lossy(&output.stderr));
        let result: String = "error".to_string();
        return result
    }
}

fn render_page(url: Url, source: bool) -> Vec<String> {
	clear_screen();
	let mut content = fetch_page(&url);
	let mut links = Vec::new();
	let (screen_width, _screen_height) = termion::terminal_size().unwrap();

	if source == true {
		content += &format!("{}", &"Viewing source code".yellow());
	}
	else if &content[..13] == "<!DOCTYPE md>" {
		(content, links) = parse_markdown((&content[13..]).to_string());
	}
	else {
		content += &format!("{}", &"Warning: This page is invalid markdown, it should contain <!DOCTYPE md> at the very start of the file, showing raw text".yellow());
	}

	for _i in 0..screen_width {
		print!("—"); 
	}
	print!("{}\n", url);
	for _i in 0..screen_width {
		print!("—"); 
	}
	println!("\n\n{}", content);
	for _i in 0..screen_width {
		print!("—"); 
	}
	println!();

	// Return links (you can add link parsing logic)
	return links;   
}

fn input() -> String{
    let mut s=String::new();
    let _=stdout().flush();
    stdin().read_line(&mut s).expect("Did not enter a correct string");
    if let Some('\n')=s.chars().next_back() {
        s.pop();
    }
    if let Some('\r')=s.chars().next_back() {
        s.pop();
    }
    return s;
}

fn parse_url(user_input: String, previous_url: &Url) -> Result<Url, ParseError> {
	println!("user input: {}",user_input);
	println!("previous url: {:?}",previous_url);
	let to_parse =  if user_input.contains("://")  {
		println!("Contains different scheme or is a path");
		user_input
	}
	else if user_input[..1] == *"/" {
		format!("http://{}/{}",Url::host_str(previous_url).expect("ivalid").to_string(), user_input)
	}
	else {
		println!("prepending scheme to user input");
		format!("http://{}", user_input)  // Prepend 'mttp://' if no scheme is found
	};

	println!("Parsing: {}", to_parse);
	if let Ok(mut url) = Url::parse(&to_parse) {
		if url.port() == None {
			let _ = url.set_port(Some(3477));
		}
		println!("{:?}",url);
		println!("{}",url.as_str());
		println!("parsed successfully");
		return Ok(url)
	}
	else {
		return Err(ParseError::InvalidDomainCharacter)
	}

}

fn main() {
	clear_screen();
	println!("Enter a url: ");
	let user_input = input();

	if user_input == "q" {
		std::process::exit(0);
	}
	let mut load_page: bool = true;
	let mut history: Vec<Url> = Vec::new();
	let mut historical_position: usize = 0;
	let mut links: Vec<String> = Vec::new();
	let mut source: bool = false; // Wether to view source of markdown page or rendered version
	if let Ok(mut url) = parse_url(user_input, &Url::parse(&"http://deadvey.com").unwrap()) { // Change this and make internal pages ;)
		history.push(url.clone());
		'mainloop: loop {
			if load_page {
				links = render_page(history[historical_position].clone(), source);
				println!("Enter reference number to follow, h for help, or q to quit");
			}
			load_page = false;

			let user_input = input(); 
			if user_input == "q" {
				break 'mainloop;
			}
			else if user_input == "r" {
				load_page = true;
				continue;
			}
			else if user_input == "s" {
				source = ! source; // Flip the boolean to toggle source mode
				load_page = true;
			}
			else if user_input == "i" {
				let _ = url.set_path("/"); 				
				for _i in historical_position+1..history.len() {
					history.remove(historical_position+1);
				}
				history.push(url.clone());
				historical_position += 1;
				load_page = true;
			}
			else if user_input == "b" {
				if historical_position > 0 {
					historical_position -= 1;
					load_page = true;
				}
				else {
					println!("At start of history");
				}
			}
			else if user_input == "f" {
				if historical_position < history.len()-1 {
					historical_position += 1;
					load_page = true;
				}
				else {
					println!("At end of history");
				}
			}
			else if user_input == "h" {
				println!("Source code: https://git.javalsai.dynv6.net/deadvey/markdown-webbrowser
q: quit
h: help
r: reload
s: view source code of page
i: visit root index of this host eg: root index of mttp://deadvey.com/blog/4.md is just deadvey.com
b: go back in history
f: go forward in history
ox: print the hyprlink of reference x eg: o5 or o24
[url]: follow the inputed url");
			}
			else if user_input.chars().nth(0).unwrap() == 'o' {
			    let number_str = &user_input[1..];
			    if let Ok(number) =  number_str.parse::<usize>() {
				println!("{}", links[number]);
			    } else {
				println!("error");
			    }
			}
			else if let Ok(number) = user_input.parse::<usize>() {
				if number < links.len() {
					if let Ok(parsed_value) = parse_url(links[number].clone(), &url.clone()) {
						url = parsed_value;
						for _i in historical_position+1..history.len() {
							history.remove(historical_position+1);
						}
						history.push(url.clone());
						historical_position += 1;
						load_page = true;
					}
					else {
						println!("Invalid url\nAttempting to open url in web browser");
					}
				} else {
					println!("Invalid reference id");
				}
			}
			else if let Ok(parsed_value) = parse_url(user_input, &url.clone()) {
				url = parsed_value;
				for _i in historical_position+1..history.len() {
					history.remove(historical_position+1);
				}
				history.push(url.clone());
				historical_position += 1;
				load_page = true;
			}
			else {
				println!("Invalid input");
			}
		}
	}
	else {
		println!("Invalid mttp url, try mttp:// at the start of your input.");
	}
}

